<?
/************************************************************************************
simpleparser.php - Parse HTML into a stream of nodes.
Copyright (C) 2006  Daniel Rairigh

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

You can contact the Author by email at WebMaster@drlock.com

 ************************************************************************************/

/*
* This library was orginally inspired by phphtmlparser found at: 
* Original code from: http://php-html.sourceforge.net/
*
* CHANGE LOG:
* Oct 22, 2008 -- Accepted commented out tags such as "<!--p></p-->" as valid comments.
*		  Patch contributed by Roman Kirillov http://sigizmund.com
*/
?>
<?php

define ("NODE_TYPE_START",      0);
define ("NODE_TYPE_ELEMENT",    1);
define ("NODE_TYPE_ENDELEMENT", 2);
define ("NODE_TYPE_TEXT",       3);
define ("NODE_TYPE_COMMENT",    4);
define ("NODE_TYPE_DONE",       5);

class SimpleParser {
        
        var $version = 0.01; //Alpha release

	/******************
	* PUBLIC Properties
	*******************/
	var $NodeType; /* one of the NODE_TYPE_* constants */
	
	/*
	* For NodeType NODE_TYPE_ELEMENT this is True if tag has closing slash.
	*   ie. <img alt="" src="" />
	*   (Note: img, bg are always self closing even if not explicit in HTML,
	*    see $alwaysSelfClosing)
	* In all other cases this is false
	*/
	var $NodeSelfClosing;
    
	/* Name of tag (always lowercase), 'Text', or 'Comment' */
	var $NodeName = "";

	/* Hash of all tag attributes.  Attribute names (hash key) are always lowercase */
	var $NodeAttributes;

	/*
	* How many ancestors this tag has.
	* Opening and closing tags, should have the same nesting level.
	* (If not a non-fatal error will be issued).
	* $NodeNestingDepth is always -1 for node types other than NODE_TYPE_ELEMENT &
	* NODE_TYPE_ENDELEMENT  
	*/
	var $NodeNestingDepth;

	/* Character Data: for Text & Comment nodes */
	var $CData = "";
	
	/* Line number the current Node starts on */
	var $LineNumber;

	/* Array of all errors encountered */
	var $Errors = array();

	/******************
	* PRIVATE Properties
	*******************/
	var $HtmlElements; //Array of html split on < & >
	var $HtmlArrayLength; //Always equal to sizeof($HtmlElements), used only for speed
	
	var $CurrentNodeIndex; //Index of current element
	var $LineNumberCount; //Number of carrage returns we have seen so far
	var $tagStack = array(); //Current nesting level stack
	
	//Tags that are always self closing, even without the ending slash
	var $alwaysSelfClosing = array('img'=>1, 'br'=>1, 'input'=>1, 'link'=>1, 'hr'=>1, 'meta'=>1);
    
	/******************
	* PUBLIC Methods
	*******************/
	
	/* SimpleParser - Contructor
	 * Args: HtmlText - String with HTML to be parsed
	 * Return: SimpleParser object
	 */
	function SimpleParser ($HtmlText) {
		$this->CurrentNodeIndex = -1; //No Node current yet
		$this->clearAttributes();
		$this->tokenize($HtmlText);
		$this->LineNumberCount = 1;
		
		$this->NodeType = NODE_TYPE_START;
	}
    
	
	/* nextNode - Parse the next node in the HTML
	 * Args: None
	 * Return: False on End of HTML or fatal error, else True
	 */
	function nextNode() {
		do {
			$token = $this->nextToken();
			if (! $token) { $this->NodeType = NODE_TYPE_DONE; return false; }
			
			$result = $this->readTag($token);
			
		//Skip over empty text nodes
		} while ($result && $this->NodeType == NODE_TYPE_TEXT && ! $this->CData);
		return $result;
	}
	
	/* isValidTagIdentifier - Test for valid tag name
	 * Args: Tag name
	 * Return: True if tag name is valid, else false
	 */
	function isValidTagIdentifier ($name) {
		return ereg ("^[A-Za-z0-9_\\-]+$", $name);
	}

	/* clearErrors - Clear previous errors 
	 * Args: None
	 * Return: None
	 */
	function clearErrors() {
		$this->Errors = array();
	}
        
        
	/* formatHTML - Formats current node for output 
	 * Args: indent - OPTIONAL if not NULL contents of indent are used to indent tags (try "\t" or " ")
	 * Return: Formatted HTML
         * TODO: This function is completely un-tested
	 */
        function formatHTML($indent = NULL) {
            if ($this->NodeType) {
                return $this->CData;
            }
            
            $rtn = "";
            if (! is_null($indent)) { $rtn .= str_repeat($indent, $this->NodeNestingDepth); }
            $rtn .= "<{$this->NodeName}";
            if ($this->NodeAttributes) {
                foreach($this->NodeAttributes AS $key => $value) {
                    $rtn .= " $key=\"" . addslashes($value) . "\"";
                }
            }
            if ($this->NodeSelfClosing) { $rtn .= " /"; }
            $rtn .= ">\n";
            return $rtn;
        }

	/******************
	* PRIVATE Methods
	*******************/

	/* tokenize - Break HTML string into useful tokens 
	 * Args: HtmlText - string of HTML to be tokenized
	 * Return: None
	 */
	function tokenize($HtmlText) {
		/**
		 * There are at least 4 ways to do this: (tested with a 290k string)
		 * strtok()     :: 0.0553269386292 sec
		 * explode()    :: 0.2658162117 sec
		 * preg_split() :: 0.367875099182 sec
		 * split()      :: 15.1390748024 sec
		 * 
		 * preg_split() returns the most useful information, because I can leave the "<" and ">"
		 * in the tokens, so I know which was which.  Also, we can exlude empty tokens
		 * Extending strtok to differentiate between < and >, raises the run time to 0.2106
		 */
		
		//This will keep the < & > at the beging of each item, so we can tell
		//what type of string it was. Also we drop all empty tokens
		$this->HtmlElements = preg_split("/(?=[<>])/",$HtmlText,-1,PREG_SPLIT_NO_EMPTY);
		$this->HtmlArrayLength = sizeof($this->HtmlElements);
	}

	/* readTag - Parse an token from the HTML string
	* Args: Text of an html token (Must begin with < or >)
	* return: false on fatal error else, true
	*/
	function readTag($text) {
		//Break node into key parts
		if (! preg_match("!^([<>])(/?)(.*)!s", $text, $match) ) {
		    $this->setError("Unrecognized token: $text");
		    $this->NodeType = NODE_TYPE_DONE;
		    return false;
		}
		list(/*NULL*/,$type_flag, $end_flag, $content) = $match;
		
		//Initialize node properties
		$this->clearAttributes();
		$this->NodeSelfClosing = false;
		$this->CData = "";
		$trailingSlash = false;
		
		//Process the tag
		if ($type_flag == '<') {
			if ( substr($content,-1) == "/") { //This tag is self closing
				$trailingSlash = true;
				$content = substr($content,0, strlen($content)-1);
			}
			
			//Begin spliting $content on space
			if (! preg_match('/^(\S*)\s*(.*)$/s', $content, $matches)) {
				$this->setError("Internal error, this string did not match tag name RE in readTag: $content");
			}
			
			$this->NodeName = strtolower($matches[1]);
			$remainder = $matches[2];
			
			if ( substr($this->NodeName, 0, 3) == '!--') { //PHP gracefully handles strings less than 3 chars
			    //HTML Comment
				$this->NodeType = NODE_TYPE_COMMENT;
				$this->NodeName = "Comment";
				$this->CData = $this->findCommentEnd(true);
				return true;
			} else if ($this->NodeName == '![cdata[') {
			    //Character Data
				$this->NodeType = NODE_TYPE_TEXT;
				$this->NodeName = "CDATA"; //NOTE: This is NOT Text
				$this->CData = $this->findCommentEnd(true);
				return true;
			} else if (! $this->isValidTagIdentifier($this->NodeName) ) {
				//Not a valid tag, treat it as text
				$type_flag = ">";
				$content = "<" . $content;
			} else {
				//Parse all other attributes
				while (preg_match('/^(\S+)\s*\=\s*((?:"[^"]*")|(?:\'[^\']*\'))\s*(.*)$/',$remainder,$matches)) {
					$attrName = $matches[1];
					$value = $matches[2];
					$remainder = $matches[3];
					$this->NodeAttributes[strtolower($attrName)] = $this->unquote($value);
				}
			}
		}
		
		//Set node properties
		if ($type_flag == '>') {
		    $this->NodeType = NODE_TYPE_TEXT;
		    $this->NodeName = "Text";
		    $this->CData = $end_flag . $content;
		} else if ($end_flag == "/") {
		    $this->NodeType = NODE_TYPE_ENDELEMENT;
		} else {
		    $this->NodeType = NODE_TYPE_ELEMENT;
		    if ( $trailingSlash || $this->alwaysSelfClosing[$this->NodeName] ) {
			    $this->NodeSelfClosing = true;
		    }
		//TODO: Deal with Script tags which have HTML inside of them - like findCommentEnd()
		}
		$this->NodeNestingDepth = $this->calcTagDepth();
		
		return true;            
	}
    
	/* calcTagDepth - Calculate nesting depth of the tag, this also raises nesting errors
	* Args: None
	* Return: Tag depth
	*/
	function calcTagDepth() {
	    if ($this->NodeType == NODE_TYPE_ELEMENT) {
		    if ($this->NodeSelfClosing) {
			    //There is no point in pushing and immediately poping this tag
			    $offset = +1;
		    } else {
			    $this->tagStack[] = $this->NodeName;
			    $offset = 0;
		    }
	    } else if ($this->NodeType == NODE_TYPE_ENDELEMENT) {
		    $tag = array_pop($this->tagStack);
		    if ($tag != $this->NodeName) {
			    $this->setError(
			    "Found closing '{$this->NodeName}', expected closing '$tag'");
		    }
		    $offset = +1;
	    } else {
		    return -1;
	    }
	    return sizeof($this->tagStack) + $offset;
	}
    
	/* findCommentEnd - Assuming the node at $index is part of an HTML comment or CDATA
	* 			scan ahead and find the end of the comment or CDATA.
	* Args: $updateCurNode - True to change $this->CurrentNodeIndex to point to end of comment
	*       $index - Index of node to scan, -1 means use $this->CurrentNodeIndex
	* Return: Contents of comment
	*/
	function findCommentEnd($updateCurNode=false, $index=-1) {
		if ($index == -1) {
			$index = $this->CurrentNodeIndex;
		} else if ($updateCurNode) {
			//If this is not the Current Node, but I will be changing CurrentNode
			//Then I need to keep track of carrage returns I am passing.
			$this->aggregateLineNumbers($index);
		}
		//$index must be one less than last index of $this->HtmlArrayLength
		if ($index > $this->HtmlArrayLength-1) {
			$this->setError("End of HTML reached without finding end of comment.");
			return '';
		}
		
		$content = $this->HtmlElements[$index];
		$end_of_str = substr($content, -2);  //Looking for '--' or ']]' preceding '>'
		$next_str_flag = substr($this->HtmlElements[$index+1],0,1); //Looking for '>'
		
		//<!-- Comment --> <![CDATA[ Character Data ]]>
		if (($end_of_str == "--" || $end_of_str == "]]") && $next_str_flag == ">") {
		    //End of comment found
			//NOTE: This will break on nested HTML comments,
			//but I think that is invalid anyway
			if ($updateCurNode) { $this->CurrentNodeIndex = $index; }
			$content = $this->HtmlElements[$index] . '>';
			return $content;
		} else {
			return $content . $this->findCommentEnd($updateCurNode, $index+1);
		}
	}

	/* nextToken - Move to and return the next elment in HTML array
	* Args: None
	* Return: current token text
	*/
	function nextToken() {
		//Indexing a large array, seems to be WAY faster than shifting off of it
		$this->CurrentNodeIndex++;
		if ($this->CurrentNodeIndex >= $this->HtmlArrayLength) {
			$this->CurrentNodeIndex = -1;
			$this->LineNumberCount = 1;
			return false;
		}
		
		$this->LineNumber = $this->aggregateLineNumbers($this->CurrentNodeIndex);
		return $this->HtmlElements[$this->CurrentNodeIndex];
	}
	
	/* aggregateLineNumbers - Add the carrage return count at $index to the total line number
	 *			  count.
	 * Args: index - array index of token to read
	 * Return: Total line numbers before $index was added
	 */
	function aggregateLineNumbers($index) {
		$prev_cnt = $this->LineNumberCount;
		
		$token = $this->HtmlElements[$index];
		$this->LineNumberCount += 
				max(substr_count( $token, "\n"), substr_count( $token, "\r"));
		return $prev_cnt;
	}

	/* setError - Log an error message in the error queue
	* Args: Error message
	* Return: false
	*/
	function setError($msg) {
		$this->Errors[] = $msg . 
			" Line: {$this->LineNumber}" .
			" BACKTRACE: " . implode("/", $this->tagStack);
		return $tokenfalse;
	}

	/* clearAttributes - Reset all attributes for the new node
	 * Args: None
	 * Return: None
	 */
	function clearAttributes() {
		$this->NodeAttributes = array();
	}
    
	/* unquote - Remove quotes from a string
	* Args: String
	* Return: UnQuoted string
	*/
	function unquote($text) {
	    $text = trim($text);
	    if (preg_match("!['\"](.*)['\"]!", $text, $match) ) {
		    return $match[1];
	    } else {
		    return $text;
	    }
	}
}

?>
